(*
 * F-ing Modules with Coq
 *
 * (c) 2009 Andreas Rossberg & Claudio Russo
 *)

Require Import Metatheory.
Require Import AdditionalTactics.
Require Import Misc.
Require Import Fomega.
Require Import Derived_multiple.
Require Import Derived_env.


(*************************************************************************)
(* N-ary functions *)
(*************************************************************************)

(* t1 .. tN -> t   :=   t1 -> ... -> tN -> t *)
Fixpoint arrsT ts t {struct ts} : typ :=
  match ts with
  | nil => t
  | t'::ts' => arrT t' (arrsT ts' t)
  end.

(* \x1 : t1 ... xN : tN. e   :=   \x1 : t1. ... . \xN : tN. e *)
Fixpoint lamsE ts e {struct ts} : exp :=
  match ts with
  | nil => e
  | t'::ts' => lamE t' (lamsE ts' e)
  end.

(* e e1 ... eN   :=   e e1 ... eN *)
Fixpoint appsE e es {struct es} : exp :=
  match es with
  | nil => e
  | e'::es' => (appsE (appE e e') es')
  end.


(*************************************************************************)
(* Substitution *)
(*************************************************************************)

Lemma substAT_arrs : forall a t' ts t,
  substAT a t' (arrsT ts t) = arrsT (map (substAT a t') ts) (substAT a t' t).
Proof.
  intros a t' ts t.
  generalize dependent t.
  induction ts; intros t; simpl arrsT; simpl substAT.
    Case "ts = nil".
      reflexivity.
    Case "ts = t::ts'".
      rewrite IHts. reflexivity.
Qed.

Corollary openAT_arrs : forall t' ts t,
  openAT t' (arrsT ts t) = arrsT (map (openAT t') ts) (openAT t' t).
Proof.
  intros t' ts t. unfold openAT.
  rewrite substAT_arrs. unfold substAT. reflexivity.
Qed.


Lemma substAE_lams : forall a t ts e,
  substAE a t (lamsE ts e) = lamsE (map (substAT a t) ts) (substAE a t e).
Proof.
  intros a t ts e.
  generalize dependent e.
  induction ts; intros e; simpl lamsE; simpl substAE; simpl substAT.
    Case "ts = nil".
      reflexivity.
    Case "ts = t::ts'".
      rewrite IHts. reflexivity.
Qed.

Corollary openAE_lams : forall t ts e,
  openAE t (lamsE ts e) = lamsE (map (openAT t) ts) (openAE t e).
Proof.
  intros t ts e. unfold openAE.
  rewrite substAE_lams. unfold openAT. unfold substAT. reflexivity.
Qed.

Lemma substXE_lams : forall x e' ts e,
  substXE x e' (lamsE ts e) = lamsE ts (substXE (shifts (length ts) x) e' e).
Proof.
  intros x e' ts e.
  generalize dependent x. generalize dependent e.
  induction ts; intros e x; simpl lamsE; simpl substXE.
    Case "ts = nil".
      reflexivity.
    Case "ts = t::ts'".
      rewrite IHts. rewrite shift_shifts. rewrite shifts_shift. reflexivity.
Qed.

Corollary openXE_lams : forall e' ts e,
  openXE e' (lamsE ts e) = lamsE ts (substXE (length ts) e' e).
Proof.
  intros e' ts e. unfold openXE.
  rewrite substXE_lams. rewrite shifts_bnd. rewrite plus_zero. reflexivity.
Qed.


Lemma substAE_apps : forall a t e es,
  substAE a t (appsE e es) = appsE (substAE a t e) (map (substAE a t) es).
Proof.
  intros a t e es.
  generalize dependent e.
  induction es; intros e; simpl appsE; simpl substAE.
    Case "es = nil".
      reflexivity.
    Case "es = e'::es'".
      rewrite IHes. simpl substAE. reflexivity.
Qed.

Corollary openAE_apps : forall t e es,
  openAE t (appsE e es) = appsE (openAE t e) (map (openAE t) es).
Proof.
  intros t e es. unfold openAE.
  rewrite substAE_apps. unfold substAE. reflexivity.
Qed.


Lemma substXE_apps : forall x e' e es,
  substXE x e' (appsE e es) = appsE (substXE x e' e) (map (substXE x e') es).
Proof.
  intros x e' e es.
  generalize dependent e.
  induction es; intros e; simpl appsE; simpl substXE.
    Case "es = nil".
      reflexivity.
    Case "es = e'::es'".
      rewrite IHes. simpl substXE. reflexivity.
Qed.

Corollary openXE_apps : forall e' e es,
  openXE e' (appsE e es) = appsE (openXE e' e) (map (openXE e') es).
Proof.
  intros e' e es. unfold openXE.
  rewrite substXE_apps. unfold substXE. reflexivity.
Qed.


(*************************************************************************)
(* Multiple substitution *)
(*************************************************************************)

Lemma substsAE_lams : forall as' ts' ts e,
  substsAE as' ts' (lamsE ts e) = lamsE (map (substsAT as' ts') ts) (substsAE as' ts' e).
Proof.
  intros as' ts'.
  induction ts; intros e; simpl lamsE; simpl substAE; simpl substAT.
    Case "ts = nil".
      reflexivity.
    Case "ts = t::ts'".
      rewrite substsAE_lam. rewrite IHts. reflexivity.
Qed.

Corollary opensAE_lams : forall ts' ts e,
  opensAE ts' (lamsE ts e) = lamsE (map (opensAT ts') ts) (opensAE ts' e).
Proof.
  intros ts' ts e. rewrite unfold_opensAE.
  rewrite substsAE_lams. rewrite unfold_opensAE. f_equal.
  induction ts; simpl map.
    reflexivity.
    rewrite unfold_opensAT. rewrite IHts. reflexivity.
Qed.

Lemma substsAE_apps : forall as' ts es e,
  substsAE as' ts (appsE e es) = appsE (substsAE as' ts e) (map (substsAE as' ts) es).
Proof.
  intros as' ts.
  induction es; intros e; simpl appsE; simpl substsAE.
    Case "es = nil".
      reflexivity.
    Case "es = e'::es'".
      rewrite IHes. rewrite substsAE_app. reflexivity.
Qed.

Corollary opensAE_apps : forall ts e es,
  opensAE ts (appsE e es) = appsE (opensAE ts e) (map (opensAE ts) es).
Proof.
  intros ts e es. rewrite unfold_opensAE.
  rewrite substsAE_apps. rewrite unfold_opensAE. f_equal.
  induction es; simpl map.
    reflexivity.
    rewrite unfold_opensAE. rewrite IHes. reflexivity.
Qed.


(*************************************************************************)
(* Local closure *)
(*************************************************************************)

Lemma lclosedT_arrs : forall ts t,
  lclosedTs ts ->
  lclosedT t ->
  lclosedT (arrsT ts t).
Proof.
  induction ts; intros t Hts Ht; simpl arrsT.
    apply Ht.
    apply lclosedT_arr.
      eapply lclosedTs_head. apply Hts.
      apply IHts.
        eapply lclosedTs_tail. apply Hts.
        apply Ht.
Qed.

Lemma lclosedE_lams : forall L ts e,
  lclosedTs ts ->
  (forall xs, fresh L (length ts) xs -> lclosedE (opensXE (freeEs xs) e)) ->
  lclosedE (lamsE ts e).
Proof.
  intros L ts. generalize dependent L.
  induction ts; intros L e Hts He; simpl lamsE.
    replace e with (opensXE (freeEs nil) e).
      apply He. simpl fresh. trivial.
      unfold freeEs. simpl map. simpl opensXE. reflexivity.
    apply lclosedE_lam with (L := L).
      eapply lclosedTs_head. apply Hts.
      intros x Fr.
        rewrite openXE_lams. apply IHts with (L := L \u {{x}}).
          eapply lclosedTs_tail. apply Hts.
          intros xs Frs.
            replace (length ts) with (length (freeEs xs)).
              rewrite <- opensXE_cons.
                rewrite <- freeEs_cons. apply He. simpl fresh. split. apply Fr. apply Frs.
                apply lclosedE_var.
                apply lclosedEs_vars. rewrite freeEs_length. symmetry. eapply fresh_length. apply Frs.
Qed.

Lemma lclosedE_apps : forall e es,
  lclosedE e ->
  lclosedEs es ->
  lclosedE (appsE e es).
Proof.
  intros e es. generalize dependent e.
  induction es; intros e He Hes; simpl appsE.
    apply He.
    apply IHes.
      apply lclosedE_app.
        apply He.
        eapply lclosedEs_head. apply Hes.
      eapply lclosedEs_tail. apply Hes.
Qed.


(*************************************************************************)
(* Typing *)
(*************************************************************************)

Lemma ofT_arrs : forall D ts t,
  (forall t, In t ts -> ofT D t omK) ->
  ofT D t omK ->
  ofT D (arrsT ts t) omK.
Proof.
  intros D ts t Hts Ht.
  induction ts; simpl arrsT in *.
    Case "ts = nil".
      apply Ht.
    Case "ts = t' :: ts'".
      apply ofT_arr.
        apply Hts. apply in_eq.
        apply IHts. intros t0 Ht0. apply Hts. apply in_cons. apply Ht0.
Qed.

Lemma eqT_arrs : forall ts t ts' t',
  length ts = length ts' ->
  (forall ti ti', In (ti, ti') (combine ts ts') -> eqT ti ti') ->
  eqT t t' ->
  eqT (arrsT ts t) (arrsT ts' t').
Proof.
  induction ts; intros t ts' t' Len Hts Ht.
    Case "ts = nil".
      destruct ts'; simpl arrsT.
        SCase "ts' = nil".
          apply Ht.
        SCase "ts' <> nil".
          simpl length in Len. inversion Len.
    Case "ts = t::ts".
      destruct ts'; simpl arrsT.
        SCase "ts' = nil".
          simpl length in Len. inversion Len.
        SCase "ts' = t'::ts'".
          apply eqT_arr.
            apply Hts. simpl combine. apply in_eq.
            apply IHts.
              simpl in Len. inversion Len. reflexivity.
              intros ti ti' Elem. apply Hts. simpl combine. apply in_cons. apply Elem.
              apply Ht.
Qed.


Lemma ofE_lams : forall L D G ts e t,
  (forall t, In t ts -> ofT D t omK) ->
  (forall xs, fresh L (length ts) xs -> ofE D (G & xs~~ts) (opensXE (freeEs xs) e) t) ->
  ofE D G (lamsE ts e) (arrsT ts t).
Proof.
  intros L D G ts e t.
  generalize dependent t. generalize dependent e.
  generalize dependent G. generalize dependent L.
  induction ts; intros L G e t Hts He; simpl lamsE in *; simpl arrsT in *.
    Case "ts = nil".
      assert (He' : ofE D (G & nil~~nil) (opensXE (freeEs nil) e) t).
        apply He. simpl fresh. trivial.
      apply He'.
    Case "ts = t' :: ts'".
      apply ofE_lam with (L := L).
        apply Hts. apply in_eq.
        intros x Fr. rewrite openXE_lams. apply IHts with (L := L \u {{x}}).
          intros t0 Elem0. apply Hts. apply in_cons. apply Elem0.
          intros xs Frs.
            rewrite concat_assoc. rewrite <- concat_multi_cons.
            replace (length ts) with (length xs).
              rewrite <- opensXE_cons_vars. apply He.
                simpl fresh. split. apply Fr. apply Frs.
              symmetry. eapply fresh_length. apply Frs.
Qed.

Lemma ofE_apps : forall D G e es ts t,
  ofE D G e (arrsT ts t) ->
  length es = length ts ->
  (forall e' t', In (e', t') (combine es ts) -> ofE D G e' t') ->
  ofE D G (appsE e es) t.
Proof.
  intros D G e es ts t.
  generalize dependent ts. generalize dependent t. generalize dependent e.
  induction es; intros e t ts He Len Hes; simpl appsE in *.
    Case "es = nil".
      destruct ts; simpl arrsT in *; simpl length in Len.
        SCase "ts = nil".
          apply He.
        SCase "ts = t'::ts'".
          inversion Len.
    Case "es = e'::es'".
      destruct ts; simpl arrsT in *; simpl length in Len.
        SCase "ts = nil".
          inversion Len.
        SCase "ts = t'::ts'".
          inversion Len.
          apply IHes with (ts := ts).
            eapply ofE_app.
              apply He.
              apply Hes. simpl combine. apply in_eq.
            apply H0.
            intros e' t' Elem. apply Hes. simpl combine. apply in_cons. apply Elem.
Qed.
