(*
 * F-ing Modules with Coq
 *
 * (c) 2009 Andreas Rossberg & Claudio Russo
 *)

Require Import Metatheory.
Require Import Metatheory_Misc.
Require Import AdditionalTactics.
Require Import Fomega_syntax.
Require Import Fomega_vars.


(*************************************************************************)
(* Environments *)
(*************************************************************************)

Lemma empty_concat : forall (A : Set) (E : env A),
  E = empty & E.
Proof.
  intros A E. unfold empty, concat. rewrite <- app_nil_end. reflexivity.
Qed.


(*************************************************************************)
(* Variable shifting *)
(*************************************************************************)

Lemma shift_nonzero : forall x,
  shift x <> 0.
Proof.
  unfold shift.
  intros x Eq.
  destruct x; inversion Eq.
Qed.

Lemma shift_injectivity : forall x1 x2,
  shift x1 = shift x2 ->
  x1 = x2.
Proof.
  unfold shift.
  intros x1 x2 H.
  destruct x1; destruct x2.
    apply H.
    inversion H.
    inversion H.
    congruence.
Qed.

Corollary shift_injectivity_inv : forall x1 x2,
  x1 <> x2 ->
  shift x1 <> shift x2.
Proof.
  unfold not. intros. apply H. apply shift_injectivity. apply H0.
Qed.

Lemma shift_ne : forall x1 x2,
  shift x1 <> shift x2 ->
  x1 <> x2.
Proof.
  unfold not. intros. apply H. f_equal. apply H0.
Qed.


(*************************************************************************)
(* Trivial lemmas for rewriting *)
(*************************************************************************)

Lemma substAT_var : forall a t,
  substAT a t a = t.
Proof.
  intros. simpl substAT. destruct (eqOcc a a). reflexivity. destruct n. reflexivity.
Qed.

Lemma substAT_var_ne : forall a1 t a2,
  a2 <> a1 ->
  substAT a1 t a2 = a2.
Proof.
  intros. simpl substAT. destruct (eqOcc a2 a1). destruct (H e). reflexivity.
Qed.

Lemma substAT_var_gt : forall n1 n2 t,
  n1 > n2 ->
  substAT n1 t n2 = n2.
Proof.
  intros. simpl substAT. destruct (eqOcc n2 n1). inversion e. subst.
    absurd (n1 > n1).  omega. apply H. destruct (n2 === n1).
      destruct n. f_equal. apply e.
      reflexivity.
Qed.

Lemma substAT_app : forall a t t1 t2,
  substAT a t (appT t1 t2) = appT (substAT a t t1) (substAT a t t2).
Proof.
  intros. simpl substAT. reflexivity.
Qed.


Lemma substXE_var : forall x e,
  substXE x e x = e.
Proof.
  intros. simpl substXE. destruct (eqOcc x x). reflexivity. destruct n. reflexivity.
Qed.

Lemma substXE_var_ne : forall x1 e x2,
  x2 <> x1 ->
  substXE x1 e x2 = x2.
Proof.
  intros. simpl substXE. destruct (eqOcc x2 x1). destruct (H e0). reflexivity.
Qed.

Lemma substXE_var_gt : forall n1 n2 e,
  n1 > n2 ->
  substXE n1 e n2 = n2.
Proof.
  intros. simpl substXE. destruct (eqOcc n2 n1). inversion e0. subst.
    absurd (n1 > n1).  omega. apply H. destruct (n2 === n1).
      destruct n. f_equal. apply e0.
      reflexivity.
Qed.


Lemma map_concat_single : forall G x t (f : typ -> typ),
  Env.map f (G & x~t) = Env.map f G & x~(f t).
Proof.
  intros G x t f. simpl. reflexivity.
Qed.


(*************************************************************************)
(* Substituting fresh variables *)
(*************************************************************************)

Lemma substAT_fresh : forall a t' t,
  a \notin varsAT t ->
  substAT a t' t = t.
Proof.
  intros a t' t Ha.
  induction t; simpl in *; f_equal; eauto.
    Case "var".
      destruct (eqOcc o a); subst.
        destruct Ha. apply S.singleton_2. reflexivity.
        reflexivity.
Qed.

Lemma substAE_fresh : forall a t' e,
  a \notin varsAE e ->
  substAE a t' e = e.
Proof.
  intros a t' e Ha.
  induction e; simpl in *; f_equal; try apply substAT_fresh; eauto.
Qed.

Lemma substXE_fresh : forall x e' e,
  x \notin varsXE e ->
  substXE x e' e = e.
Proof.
  intros x e' e Hx.
  induction e; simpl in *; f_equal; eauto.
    Case "var".
      destruct (eqOcc o x); subst.
        destruct Hx. apply S.singleton_2. reflexivity.
        reflexivity.
Qed.


Lemma substAG_fresh : forall a t' G,
  a \notin varsAG G ->
  substAG a t' G = G.
Proof.
  intros a t' G Ha.
  unfold substAG.
  unfold varsAG in Ha.
  induction G.
    Case "empty".
      simpl. reflexivity.
    Case "var".
      destruct a0.
      simpl Env.map.
      simpl fold_right in Ha.
      rewrite IHG. rewrite substAT_fresh. reflexivity.
      notin_solve. notin_solve.
Qed.


(*************************************************************************)
(* Substituting types in locally closed types *)
(*************************************************************************)

Lemma substAT_substAT_irrel : forall t a1 t1 a2 t2,
  a1 <> a2 ->
  substAT a1 t1 t = substAT a2 t2 (substAT a1 t1 t) ->
  substAT a2 t2 t = t.
Proof.
  induction t; intros a1 t1' a2 t2' Neq H; simpl in *; inversion H; f_equal;
    try solve
    [ reflexivity
    | first [eapply IHt | eapply IHt1]; [apply Neq | apply H1]
    | eapply IHt2; [apply Neq | apply H2]
    | eapply IHt with (a1 := shift a1);
        [ intro Eq; apply (Neq (shift_injectivity a1 a2 Eq))
        | apply H1 ]
    ].
    Case "var".
      destruct (eqOcc o a2); subst.
        destruct (eqOcc a2 a1).
          elim (Neq (sym_eq e)).
          simpl in H, H1. destruct (eqOcc a2 a2); subst.
            reflexivity.
            destruct n0. reflexivity.
        reflexivity.
Qed.

Lemma substAT_lclosed : forall (n : nat) t' t,
  lclosedT t ->
  substAT n t' t = t.
Proof.
  intros n t' t Ht.
  generalize dependent n.
  induction Ht; intros n; simpl substAT; f_equal;
    try solve
    [ reflexivity
    | apply IHHt
    | apply IHHt1
    | apply IHHt2
    | unfold openAT in H, H0; pick_fresh_gen L a;
      apply (substAT_substAT_irrel t2 0 a (S n) t');
        [ unfold not; intros H'; inversion H'
        | symmetry; apply H0; apply Fr ]
    ].
Qed.

Corollary openAT_lclosed : forall t' t,
  lclosedT t ->
  openAT t' t = t.
Proof.
  unfold openAT. apply substAT_lclosed.
Qed.


(*************************************************************************)
(* Substituting types in locally closed terms *)
(*************************************************************************)

Lemma substAE_substAE_irrel : forall e a1 t1 a2 t2,
  a1 <> a2 ->
  substAE a1 t1 e = substAE a2 t2 (substAE a1 t1 e) ->
  substAE a2 t2 e = e.
Proof.
  induction e; intros a1 t1' a2 t2' Neq H; simpl in *; inversion H; f_equal;
    try solve
    [ reflexivity
    | first [eapply IHe | eapply IHe1 | eapply IHe2 | eapply IHe3];
       [apply Neq
       | first [ apply H1 | apply H2 | apply H3] ]
    | eapply substAT_substAT_irrel with (a1 := a1) (a2 := a2);
       [ apply Neq
       | first [apply H1 | apply H2 | apply H3] ]
    | first [eapply IHe with (a1 := shift a1) | eapply IHe2 with (a1 := shift a1)];
        [ intro Eq; apply (Neq (shift_injectivity a1 a2 Eq))
        | first [apply H1 | apply H2] ]
    ].
Qed.

Lemma substAE_substXE_irrel : forall e x1 e1 a2 t2,
  substXE x1 e1 e = substAE a2 t2 (substXE x1 e1 e) ->
  substAE a2 t2 e = e.
Proof.
  induction e; intros x1 e1' a2 t2' H; simpl in *; inversion H; f_equal;
    try solve
    [ reflexivity
    | eapply IHe; first [apply H1 | apply H2]
    | eapply IHe1; apply H1
    | eapply IHe2; apply H2
    | eapply IHe3; apply H3
    | rewrite <- H1; symmetry; apply H1
    | rewrite <- H2; symmetry; apply H2
    | rewrite <- H3; symmetry; apply H3
    ].
Qed.

Lemma substAE_lclosed : forall (n : nat) t' e,
  lclosedE e ->
  substAE n t' e = e.
Proof.
  intros n t' e He.
  generalize dependent n.
  induction He; intros n; simpl substAE; f_equal;
    try solve
    [ reflexivity
    | apply IHHe | apply IHHe1 | apply IHHe2
    | apply substAT_lclosed; first [apply H | apply H0]
    | pick_fresh_gen L x; eapply substAE_substXE_irrel with (x1 := 0) (a2 := n);
        fold substAE in *; unfold openXE in *;
        symmetry; first [apply H0 | apply H1 | apply H2 | apply H3]; apply Fr
    ].
    Case "gen".
      unfold openAE in H0.
      pick_fresh_gen L a.
      eapply substAE_substAE_irrel with (a1 := 0) (a2 := S n).
        unfold not. intros H'. inversion H'.
        symmetry. apply H0. apply Fr.
    Case "open 2".
      unfold openAE in H0.
      pick_fresh_gen L a.
      pick_fresh_gen L x.

      eapply substAE_substAE_irrel with (a1 := 0) (a2 := S n).
        intros Eq. inversion Eq.
      symmetry.
      eapply substAE_substXE_irrel with (x1 := 0) (a2 := S n).
        fold substAE in *. fold substXE in *. unfold openXE, openAE in *.
          symmetry. apply H0. apply Fr. apply Fr0.
Qed.

Corollary openAE_lclosed : forall t' e,
  lclosedE e ->
  openAE t' e = e.
Proof.
  unfold openAE. apply substAE_lclosed.
Qed.


(*************************************************************************)
(* Substituting terms in locally closed terms *)
(*************************************************************************)

Lemma substXE_substXE_irrel : forall e x1 e1 x2 e2,
  x1 <> x2 ->
  substXE x1 e1 e = substXE x2 e2 (substXE x1 e1 e) ->
  substXE x2 e2 e = e.
Proof.
  induction e; intros x1 e1' x2 e2' Neq H; simpl in *; inversion H; f_equal;
    try solve
    [ reflexivity
    | eapply IHe; [apply Neq | apply H1]
    | eapply IHe1; [apply Neq | apply H1]
    | eapply IHe2; [apply Neq | apply H2]
    | eapply IHe3; [apply Neq | apply H3]
    | eapply IHe with (x1 := shift x1);
        [intro Eq; apply (Neq (shift_injectivity x1 x2 Eq)) | apply H1]
    | eapply IHe2 with (x1 := shift x1);
        [intro Eq; apply (Neq (shift_injectivity x1 x2 Eq)) | apply H2]
    | eapply IHe3 with (x1 := shift x1);
        [intro Eq; apply (Neq (shift_injectivity x1 x2 Eq)) | apply H3]
    ].
    Case "var".
      destruct (eqOcc o x2); subst.
        destruct (eqOcc x2 x1).
          elim (Neq (sym_eq e)).
          simpl in H. destruct (eqOcc x2 x2); subst.
            reflexivity.
            destruct n0. reflexivity.
        reflexivity.
Qed.

Lemma substXE_substAE_irrel : forall e a1 t1 x2 e2,
  substAE a1 t1 e = substXE x2 e2 (substAE a1 t1 e) ->
  substXE x2 e2 e = e.
Proof.
  induction e; intros a1 t1' x2 e2' H; simpl in *; inversion H; f_equal;
    try solve
    [ reflexivity
    | eapply IHe; apply H1
    | eapply IHe1; apply H1
    | eapply IHe2; apply H2
    | eapply IHe3; apply H3
    | eapply IHe with (x2 := shift x2); apply H1
    ].
Qed.

Lemma substXE_lclosed : forall (n : nat) e' e,
  lclosedE e ->
  substXE n e' e = e.
Proof.
  intros n e' e He.
  generalize dependent n.
  induction He; intros n; simpl substXE; f_equal;
    try solve
    [ reflexivity
    | apply IHHe
    | apply IHHe1
    | apply IHHe2
    | pick_fresh_gen L x; eapply substXE_substXE_irrel with (x1 := 0) (x2 := S n);
        [ unfold not; intros H'; inversion H'
        | symmetry; unfold openXE in *; first [apply H0 | apply H1 | apply H2]; apply Fr ]
    ].
    Case "gen".
      pick_fresh_gen L a.
      apply (substXE_substAE_irrel e2 0 a n e').
        fold substXE in *. unfold openAE in *.
        symmetry. apply H0. apply Fr.
    Case "unpack 2".
      pick_fresh_gen L a.
      pick_fresh_gen L x.
      apply (substXE_substAE_irrel e2 0 a (S n) e').
      symmetry.
      apply (substXE_substXE_irrel (substAE 0 a e2) 0 x (S n) e'). intro Eq. inversion Eq.
          fold substAE in *. fold substXE in *. unfold openXE, openAE in *.
            symmetry. apply H0. apply Fr. apply Fr0.
Qed.

Corollary openXE_lclosed : forall e' e,
  lclosedE e ->
  openXE e' e = e.
Proof.
  unfold openXE. apply substXE_lclosed.
Qed.


(*************************************************************************)
(* Opening types with types as substitution *)
(*************************************************************************)

Lemma substAT_bnd_as_substAT_free : forall a t' (n : nat) t,
  a \notin varsAT t ->
  substAT n t' t = substAT a t' (substAT n a t).
Proof.
  intros a t' n t H.
  unfold substAT.
  generalize dependent n.
  induction t; intros n; f_equal;
    try solve
    [ reflexivity
    | first [apply IHt | apply IHt1 | apply IHt2];
        unfold varsAT in H; fold varsAT in H; notin_solve
    | apply IHt with (n := S n); unfold varsAT in H; fold varsAT in H; notin_solve
    ].
    Case "var".
      destruct (eqOcc o n).
        SCase "o = a".
          unfold eqOcc. destruct (a == a). reflexivity. destruct n0. reflexivity.
        SCase "o <> a".
          destruct (eqOcc o a); subst.
            destruct H. unfold varsAT. apply S.singleton_2. reflexivity.
            reflexivity.
Qed.

Corollary openAT_as_substAT : forall a t' t,
  a \notin varsAT t ->
  openAT t' t = substAT a t' (openAT a t).
Proof.
  intros a t' t H. unfold openAT.
  apply substAT_bnd_as_substAT_free with (n := 0). apply H.
Qed.


(*************************************************************************)
(* Opening terms with types as substitution *)
(*************************************************************************)

Lemma substAE_bnd_as_substAE_free : forall a t (n : nat) e,
  a \notin varsAE e ->
  substAE n t e = substAE a t (substAE n a e).
Proof.
  intros a t n e H.
  unfold substAE.
  generalize dependent n.
  induction e; intros n; f_equal;
    try solve
    [ reflexivity
    | first [apply IHe | apply IHe1 | apply IHe2 | apply IHe3];
        unfold varsAE in H; fold varsAE in H; notin_solve
    | first [apply IHe with (n := S n) | apply IHe2 with (n := S n)];
        unfold varsAE in H; fold varsAE in H; notin_solve
    | apply substAT_bnd_as_substAT_free; unfold varsAE in H; fold varsAE in H; notin_solve
    ].
Qed.

Corollary openAE_as_substAE : forall a t e,
  a \notin varsAE e ->
  openAE t e = substAE a t (openAE a e).
Proof.
  intros a t e H. unfold openAE.
  apply substAE_bnd_as_substAE_free with (n := 0). apply H.
Qed.


(*************************************************************************)
(* Opening terms with terms as substitution *)
(*************************************************************************)

Lemma substXE_bnd_as_substXE_free : forall x e' (n : nat) e,
  x \notin varsXE e ->
  substXE n e' e = substXE x e' (substXE n x e).
Proof.
  intros x e' n e H.
  unfold substXE.
  generalize dependent n.
  induction e; intros n; f_equal;
    try solve
    [ reflexivity
    | first [apply IHe | apply IHe1 | apply IHe2];
        unfold varsXE in H; fold varsXE in H; notin_solve
    | first [apply IHe with (n := S n) | apply IHe2 with (n := S n) | apply IHe3 with (n := S n)];
        unfold varsXE in H; fold varsXE in H; notin_solve
    ].
    Case "var".
      destruct (eqOcc o n).
        SCase "o = x".
          unfold eqOcc. destruct (x == x). reflexivity. destruct n0. reflexivity.
        SCase "o <> x".
          destruct (eqOcc o x); subst.
            destruct H. unfold varsXE. apply S.singleton_2. reflexivity.
            reflexivity.
Qed.

Corollary openXE_as_substXE : forall x e' e,
  x \notin varsXE e ->
  openXE e' e = substXE x e' (openXE x e).
Proof.
  intros x e' e H. unfold openXE.
  apply substXE_bnd_as_substXE_free with (n := 0). apply H.
Qed.


(*************************************************************************)
(* Commuting independent substitutions *)
(*************************************************************************)

Definition indepAT o t :=
  match o with
  | free a => a \notin varsAT t
  | bnd n => lclosedT t
  end.

Definition indepAE o e :=
  match o with
  | free a => a \notin varsAE e
  | bnd n => lclosedE e
  end.

Definition indepXE o e :=
  match o with
  | free x => x \notin varsXE e
  | bnd n => lclosedE e
  end.

Lemma indepAT_shift : forall o t,
  indepAT o t = indepAT (shift o) t.
Proof.
  intros o t. destruct o; unfold shift.
    reflexivity.
    unfold indepAT. reflexivity.
Qed.

Lemma indepAE_shift : forall o e,
  indepAE o e = indepAE (shift o) e.
Proof.
  intros o e. destruct o; unfold shift.
    reflexivity.
    unfold indepAT. reflexivity.
Qed.

Lemma indepXE_shift : forall o e,
  indepXE o e = indepXE (shift o) e.
Proof.
  intros o e. destruct o; unfold shift.
    reflexivity.
    unfold indepXE. reflexivity.
Qed.


Lemma substAT_substAT : forall t1 t2 t a1 a2,
  a1 <> a2 ->
  indepAT a1 t2 ->
  indepAT a2 t1 ->
  substAT a1 t1 (substAT a2 t2 t) = substAT a2 t2 (substAT a1 t1 t).
Proof.
  intros t1 t2.
  induction t; intros a1 a2 Neq H1 H2; simpl substAT;
    try solve
    [ reflexivity
    | f_equal; apply IHt; [apply Neq | apply H1 | apply H2]
    | f_equal; [apply IHt1 | apply IHt2]; first [apply Neq | apply H1 | apply H2]
    | f_equal; apply IHt; [apply shift_injectivity_inv; apply Neq | rewrite <- indepAT_shift; apply H1 | rewrite <- indepAT_shift; apply H2]
    ].
      Case "var".
        destruct (eqOcc o a1); destruct (eqOcc o a2); subst.
          destruct (Neq e0).
          rewrite substAT_var. unfold indepAT in H2. destruct a2.
            rewrite substAT_fresh. reflexivity. apply H2.
            rewrite substAT_lclosed. reflexivity. apply H2.
          rewrite substAT_var. unfold indepAT in H1. destruct a1.
            rewrite substAT_fresh. reflexivity. apply H1.
            rewrite substAT_lclosed. reflexivity. apply H1.
          rewrite substAT_var_ne. rewrite substAT_var_ne.
            reflexivity.
            apply n0. apply n.
Qed.

Corollary openAT_substAT : forall t1 t2 t (a : occ),
  a <> 0 ->
  indepAT a t1 ->
  lclosedT t2 ->
  openAT t1 (substAT a t2 t) = substAT a t2 (openAT t1 t).
Proof.
  intros t1 t2 t a Neq H1 H2.
  unfold openAT.
  apply substAT_substAT.
    unfold not. intros. destruct Neq. symmetry. apply H.
    apply H2.
    apply H1.
Qed.

Lemma substAE_substAE : forall t1 t2 e a1 a2,
  a1 <> a2 ->
  indepAT a1 t2 ->
  indepAT a2 t1 ->
  substAE a1 t1 (substAE a2 t2 e) = substAE a2 t2 (substAE a1 t1 e).
Proof.
  intros t1 t2.
  induction e; intros a1 a2 Neq H1 H2; simpl substAE; f_equal;
    try solve
    [ reflexivity
    | first [apply IHe | apply IHe1 | apply IHe2 | apply IHe3 | apply substAT_substAT];
        [apply Neq | apply H1 | apply H2]
    | first [apply IHe | apply IHe1 | apply IHe2 | apply IHe3];
        [apply shift_injectivity_inv; apply Neq | rewrite <- indepAT_shift; apply H1 | rewrite <- indepAT_shift; apply H2]
    ].
Qed.

Corollary openAE_substAE : forall t1 t2 e (a : occ),
  a <> 0 ->
  indepAT a t1 ->
  lclosedT t2 ->
  openAE t1 (substAE a t2 e) = substAE a t2 (openAE t1 e).
Proof.
  intros t1 t2 e a Neq H1 H2.
  unfold openAE.
  apply substAE_substAE.
    unfold not. intros. destruct Neq. symmetry. apply H.
    apply H2.
    apply H1.
Qed.

Lemma substXE_substXE : forall e1 e2 e x1 x2,
  x1 <> x2 ->
  indepXE x1 e2 ->
  indepXE x2 e1 ->
  substXE x1 e1 (substXE x2 e2 e) = substXE x2 e2 (substXE x1 e1 e).
Proof.
  intros e1 e2.
  induction e; intros x1 x2 Neq H1 H2; simpl substXE;
    try solve
    [ reflexivity
    | f_equal; first [apply IHe | apply IHe1 | apply IHe2 | apply IHe3];
        first [ apply Neq | apply shift_injectivity_inv; apply Neq
               | apply H1 | rewrite <- indepXE_shift; apply H1
               | apply H2 | rewrite <- indepXE_shift; apply H2]
    ].
      Case "var".
        destruct (eqOcc o x1); destruct (eqOcc o x2); subst.
          destruct (Neq e0).
          rewrite substXE_var. unfold indepXE in H2. destruct x2.
            rewrite substXE_fresh. reflexivity. apply H2.
            rewrite substXE_lclosed. reflexivity. apply H2.
          rewrite substXE_var. unfold indepXE in H1. destruct x1.
            rewrite substXE_fresh. reflexivity. apply H1.
            rewrite substXE_lclosed. reflexivity. apply H1.
          rewrite substXE_var_ne. rewrite substXE_var_ne.
            reflexivity.
            apply n0. apply n.
Qed.

Corollary openXE_substXE : forall e1 e2 e (x : occ),
  x <> 0 ->
  indepXE x e1 ->
  lclosedE e2 ->
  openXE e1 (substXE x e2 e) = substXE x e2 (openXE e1 e).
Proof.
  intros e1 e2 e x Neq H1 H2.
  unfold openXE.
  apply substXE_substXE.
    unfold not. intros. destruct Neq. symmetry. apply H.
    apply H2.
    apply H1.
Qed.


Lemma substXE_substAE : forall e1 t2 e x1 a2,
  indepAE a2 e1 ->
  substXE x1 e1 (substAE a2 t2 e) = substAE a2 t2 (substXE x1 e1 e).
Proof.
  intros e1 t2.
  induction e; intros x1 a2 H; simpl substXE; simpl substAE; f_equal;
    try solve
    [ reflexivity
    | first [apply IHe | apply IHe1 | apply IHe2 | apply IHe3]; apply H
    | first [apply IHe | apply IHe1 | apply IHe2 | apply IHe3]; rewrite <- indepAE_shift; apply H
    ].
      Case "var".
        destruct (eqOcc o x1); subst.
          unfold indepAE in H. destruct a2.
            rewrite substAE_fresh. reflexivity. apply H.
            rewrite substAE_lclosed. reflexivity. apply H.
          simpl substAE. reflexivity.
Qed.

Corollary substAE_substXE : forall t1 e2 e a1 x2,
  indepAE a1 e2 ->
  substAE a1 t1 (substXE x2 e2 e) = substXE x2 e2 (substAE a1 t1 e).
Proof.
  intros. symmetry. apply substXE_substAE. apply H.
Qed.

Corollary openXE_substAE : forall e1 t2 e (a : occ),
  indepAE a e1 ->
  openXE e1 (substAE a t2 e) = substAE a t2 (openXE e1 e).
Proof.
  intros e1 t2 e a H.
  unfold openXE.
  apply substXE_substAE. apply H.
Qed.

Corollary openAE_substXE : forall t1 e2 e (x : occ),
  lclosedE e2 ->
  openAE t1 (substXE x e2 e) = substXE x e2 (openAE t1 e).
Proof.
  intros t1 e2 e x H.
  unfold openAE.
  apply substAE_substXE. apply H.
Qed.


(*************************************************************************)
(* Commuting dependent substitutions *)
(*************************************************************************)

Lemma substAT_substAT_dep : forall t1 t2 t (a1 : var) a2,
  free a1 <> a2 ->
  indepAT a2 t1 ->
  substAT a1 t1 (substAT a2 t2 t) = substAT a2 (substAT a1 t1 t2) (substAT a1 t1 t).
Proof.
  intros t1 t2 t. generalize dependent t2.
  induction t; intros t2' a1 a2 Neq H; simpl substAT;
    try solve
    [ reflexivity
    | f_equal; apply IHt; [apply Neq | apply H]
    | f_equal; [apply IHt1 | apply IHt2]; first [apply Neq | apply H]
    | f_equal; apply IHt;
        [ replace (free a1) with (shift a1); [apply shift_injectivity_inv; apply Neq | simpl shift; reflexivity]
        | rewrite <- indepAT_shift; apply H]
    ].
      Case "var".
        destruct (eqOcc o a1); destruct (eqOcc o a2); subst.
          inversion e0. destruct (Neq H0).
          rewrite substAT_var. unfold indepAT in H. destruct a2.
            rewrite substAT_fresh. reflexivity. apply H.
            rewrite substAT_lclosed. reflexivity. apply H.
          rewrite substAT_var. reflexivity.
          rewrite substAT_var_ne. rewrite substAT_var_ne.
            reflexivity.
            apply n0. apply n.
Qed.


(*************************************************************************)
(* Local closure under substitution *)
(*************************************************************************)

Lemma lclosedT_substAT : forall a t' t,
  lclosedT t' ->
  lclosedT t ->
  lclosedT (substAT a t' t).
Proof.
  intros a t' t Ht' Ht.
  generalize dependent a.
  induction Ht; intro a'; simpl substAT.
    Case "var".
      destruct a'.
        destruct (a == v). apply Ht'. apply lclosedT_var.
        apply lclosedT_var.
    Case "unit".
      apply lclosedT_unit.
    Case "arr".
      apply lclosedT_arr. apply IHHt1. apply IHHt2.
    Case "tim".
      apply lclosedT_tim. apply IHHt1. apply IHHt2.
    Case "plus".
      apply lclosedT_plus. apply IHHt1. apply IHHt2.
    Case "rec".
      apply lclosedT_rec with (L := L \u varsAT a'). intros a Fr.
      rewrite openAT_substAT.
        apply H0. notin_solve.
        apply shift_nonzero.
        destruct a'; simpl indepAT.
          apply notin_singleton_inv. simpl varsAT in Fr. notin_solve.
          apply lclosedT_var.
        apply Ht'.
    Case "all".
      apply lclosedT_all with (L := L \u varsAT a'). intros a Fr.
      rewrite openAT_substAT.
        apply H0. notin_solve.
        apply shift_nonzero.
        destruct a'; simpl indepAT.
          apply notin_singleton_inv. simpl varsAT in Fr. notin_solve.
          apply lclosedT_var.
        apply Ht'.
    Case "ex".
      apply lclosedT_ex with (L := L \u varsAT a'). intros a Fr.
      rewrite openAT_substAT.
        apply H0. notin_solve.
        apply shift_nonzero.
        destruct a'; simpl indepAT.
          apply notin_singleton_inv. simpl varsAT in Fr. notin_solve.
          apply lclosedT_var.
        apply Ht'.
    Case "triv".
      apply lclosedT_triv.
    Case "lam".
      apply lclosedT_lam with (L := L \u varsAT a'). intros a Fr.
      rewrite openAT_substAT.
        apply H0. notin_solve.
        apply shift_nonzero.
        destruct a'; simpl indepAT.
          apply notin_singleton_inv. simpl varsAT in Fr. notin_solve.
          apply lclosedT_var.
        apply Ht'.
    Case "app".
      apply lclosedT_app. apply IHHt1. apply IHHt2.
    Case "pair".
      apply lclosedT_pair. apply IHHt1. apply IHHt2.
    Case "fst".
      apply lclosedT_fst. apply IHHt.
    Case "snd".
      apply lclosedT_snd. apply IHHt.
Qed.

Corollary lclosedT_openAT : forall t' t,
  lclosedT t' ->
  lclosedT t ->
  lclosedT (openAT t' t).
Proof.
  intros t' t Ht' Ht. unfold openAT. apply lclosedT_substAT. apply Ht'. apply Ht.
Qed.


Lemma lclosedE_substAE : forall a t e,
  lclosedT t ->
  lclosedE e ->
  lclosedE (substAE a t e).
Proof.
  intros a t e Ht He.
  generalize dependent a.
  induction He; intro a; simpl substAE.
    Case "var".
      apply lclosedE_var.
    Case "triv".
      apply lclosedE_triv.
    Case "lam".
      apply lclosedE_lam with (L := L).
        apply lclosedT_substAT. apply Ht. apply H.
        intros x Fr.
          rewrite openXE_substAE.
            apply H1. notin_solve.
            destruct a; simpl indepAE. notin_solve. apply lclosedE_var.
    Case "app".
      apply lclosedE_app. apply IHHe1. apply IHHe2.
    Case "pair".
      apply lclosedE_pair. apply IHHe1. apply IHHe2.
    Case "fst".
      apply lclosedE_fst. apply IHHe.
    Case "snd".
      apply lclosedE_snd. apply IHHe.
    Case "lft".
      apply lclosedE_lft.
        apply IHHe.
        apply lclosedT_substAT. apply Ht. apply H.
    Case "rgt".
      apply lclosedE_rgt.
        apply IHHe.
        apply lclosedT_substAT. apply Ht. apply H.
    Case "case".
      apply lclosedE_case with (L := L).
        apply IHHe.
        intros x Fr.
          rewrite openXE_substAE.
            apply H0. notin_solve.
            destruct a; simpl indepAE. notin_solve. apply lclosedE_var.
        intros x Fr.
          rewrite openXE_substAE.
            apply H2. notin_solve.
            destruct a; simpl indepAE. notin_solve. apply lclosedE_var.
    Case "roll".
      apply lclosedE_roll.
        apply IHHe.
        apply lclosedT_substAT. apply Ht. apply H.
    Case "unroll".
      apply lclosedE_unroll. apply IHHe.
    Case "gen".
      apply lclosedE_gen with (L := L \u varsAT a). intros a' Fr.
      rewrite openAE_substAE.
        apply H0. notin_solve.
        apply shift_nonzero.
        destruct a; simpl indepAT.
          apply notin_singleton_inv. simpl varsAT in Fr. notin_solve.
          apply lclosedT_var.
        apply Ht.
    Case "inst".
      apply lclosedE_inst.
        apply IHHe.
        apply lclosedT_substAT. apply Ht. apply H.
    Case "pack".
      apply lclosedE_pack.
        apply lclosedT_substAT. apply Ht. apply H.
        apply IHHe.
        apply lclosedT_substAT. apply Ht. apply H0.
    Case "unpack".
      apply lclosedE_unpack with (L := L \u varsAT a).
        apply IHHe.
        intros a' x Fra' Frx.
        rewrite openAE_substAE. rewrite openXE_substAE.
          apply H0. notin_solve. notin_solve.
          destruct a; simpl indepAE. notin_solve. apply lclosedE_var.
          apply shift_nonzero.
          destruct a; simpl indepAT.
            apply notin_singleton_inv. simpl varsAT in Fra'. notin_solve.
            apply lclosedT_var.
        apply Ht.
Qed.

Corollary lclosedE_openAE : forall t e,
  lclosedT t ->
  lclosedE e ->
  lclosedE (openAE t e).
Proof.
  intros t e Ht He. unfold openAE. apply lclosedE_substAE. apply Ht. apply He.
Qed.


Lemma lclosedE_substXE : forall x e' e,
  lclosedE e' ->
  lclosedE e ->
  lclosedE (substXE x e' e).
Proof.
  intros x e' e He' He.
  generalize dependent x.
  induction He; intro x'; simpl substXE.
    Case "var".
      destruct x'.
        destruct (x == v). apply He'. apply lclosedE_var.
        apply lclosedE_var.
    Case "triv".
      apply lclosedE_triv.
    Case "lam".
      apply lclosedE_lam with (L := L \u varsXE x').
        apply H.
        intros x Fr.
          rewrite openXE_substXE.
            apply H1. notin_solve.
            apply shift_nonzero.
            destruct x'; simpl indepXE.
              apply notin_singleton_inv. simpl varsXE in Fr. notin_solve.
              apply lclosedE_var.
            apply He'.
    Case "app".
      apply lclosedE_app. apply IHHe1. apply IHHe2.
    Case "pair".
      apply lclosedE_pair. apply IHHe1. apply IHHe2.
    Case "fst".
      apply lclosedE_fst. apply IHHe.
    Case "snd".
      apply lclosedE_snd. apply IHHe.
    Case "lft".
      apply lclosedE_lft. apply IHHe. apply H.
    Case "rgt".
      apply lclosedE_rgt. apply IHHe. apply H.
    Case "case".
      apply lclosedE_case with (L := L \u varsXE x').
        apply IHHe.
        intros x Fr.
          rewrite openXE_substXE.
            apply H0. notin_solve.
            apply shift_nonzero.
            destruct x'; simpl indepXE.
              apply notin_singleton_inv. simpl varsXE in Fr. notin_solve.
              apply lclosedE_var.
            apply He'.
        intros x Fr.
          rewrite openXE_substXE.
            apply H2. notin_solve.
            apply shift_nonzero.
            destruct x'; simpl indepXE.
              apply notin_singleton_inv. simpl varsXE in Fr. notin_solve.
              apply lclosedE_var.
            apply He'.
    Case "roll".
      apply lclosedE_roll. apply IHHe. apply H.
    Case "unroll".
      apply lclosedE_unroll. apply IHHe.
    Case "gen".
      apply lclosedE_gen with (L := L). intros a Fr.
      rewrite openAE_substXE.
        apply H0. notin_solve.
        apply He'.
    Case "inst".
      apply lclosedE_inst. apply IHHe. apply H.
    Case "pack".
      apply lclosedE_pack. apply H. apply IHHe. apply H0.
    Case "unpack".
      apply lclosedE_unpack with (L := L \u varsXE x').
        apply IHHe.
        intros a x Fra Frx.
        rewrite openAE_substXE. rewrite openXE_substXE.
          apply H0. notin_solve. notin_solve.
          apply shift_nonzero.
          destruct x'; simpl indepXE.
            apply notin_singleton_inv. simpl varsXE in Frx. notin_solve.
            apply lclosedE_var.
          apply He'.
        apply He'.
Qed.

Corollary lclosedE_openXE : forall e' e,
  lclosedE e' ->
  lclosedE e ->
  lclosedE (openXE e' e).
Proof.
  intros e' e He' He. unfold openXE. apply lclosedE_substXE. apply He'. apply He.
Qed.


